unit Table;

interface

uses
  Classes, SysUtils, Generics.Collections, ComCtrls;

type
  TTableFormat = (tfExcel, tfPlain, tfCsv, tfHtml, tfListView, tfMediaWiki, tfXml);

  { TRow }

  TRow = class
    Cells: TStringList;
    procedure AddCell(Text: string);
    constructor Create;
    destructor Destroy; override;
  end;

  { TTable }

  TTable = class
  private
    function QuoteCsvString(Text: string): string;
    function ReplaceXmlEntities(Text: string): string;
  public
    Title: string;
    Columns: TStringList;
    Rows: TObjectList<TRow>;
    FirstRowIsHeader: Boolean;
    procedure Clear;
    function AddRow: TRow;
    function GetOutputTabs: string;
    function GetOutputPlain: string;
    function GetOutputCsv: string;
    function GetOutputXml: string;
    function GetOutputHtml: string;
    function GetOutputMediaWiki: string;
    procedure GetOutputListView(ListView: TListView);
    function GetOutput(OutputFormat: TTableFormat): string;
    constructor Create;
    destructor Destroy; override;
  end;

const
  TableFormatText: array[TTableFormat] of string = ('Excel paste', 'Plain text', 'CSV',
    'HTML', 'ListView', 'MediaWiki', 'XML');
  TableFormatExt: array[TTableFormat] of string = ('.txt', '.txt', '.csv',
    '.htm', '', '.txt', '.xml');


implementation

uses
  Common;

{ TTable }

function TTable.QuoteCsvString(Text: string): string;
begin
  if Text <> '' then Result := '"' + Text + '"'
    else Result := Text;
end;

function TTable.ReplaceXmlEntities(Text: string): string;
begin
  Result := StringReplace(Text, '<', '&lt;', [rfReplaceAll]);
  Result := StringReplace(Result, '>', '&gt;', [rfReplaceAll]);
  Result := StringReplace(Result, '&', '&amp;', [rfReplaceAll]);
  Result := StringReplace(Result, '''', '&apos;', [rfReplaceAll]);
  Result := StringReplace(Result, '"', '&quot;', [rfReplaceAll]);
end;

procedure TTable.Clear;
begin
  Rows.Clear;
end;

function TTable.AddRow: TRow;
begin
  Result := TRow.Create;
  Rows.Add(Result);
end;

function TTable.GetOutputTabs: string;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Rows.Count - 1 do
    Result := Result + Implode(#9, Rows[I].Cells) + LineEnding;
end;

function TTable.GetOutputPlain: string;
var
  I: Integer;
begin
  Result := '';
  for I := 0 to Rows.Count - 1 do begin
    Result := Result + Implode(LineEnding, Rows[I].Cells) + LineEnding;
    Result := Result + LineEnding;
    Result := Result + '===========================' + LineEnding;
    Result := Result + LineEnding;
  end;
end;

function TTable.GetOutputCsv: string;
var
  I: Integer;
  J: Integer;
begin
  Result := '';

  // Show columns
  for J := 0 to Columns.Count - 1 do begin
    if J > 0 then Result := Result + ',';
    if Columns[J] <> '' then
      Result := Result + QuoteCsvString(Columns[J]);
  end;
  Result := Result + LineEnding;

  // Show data rows
  for I := 0 to Rows.Count - 1 do
  with Rows[I] do begin
    for J := 0 to Cells.Count - 1 do begin
      if J > 0 then Result := Result + ',';
      if Cells[J] <> '' then
        Result := Result + QuoteCsvString(Cells[J]);
    end;
    Result := Result + LineEnding;
  end;
end;

function TTable.GetOutputXml: string;
var
  I: Integer;
  J: Integer;
begin
  Result := '<?xml version="1.0" encoding="UTF-8"?>' + LineEnding +
    '<vcard>' + LineEnding +
    '  <title>' + Title + '</title>' + LineEnding +
    '  <contacts>' + LineEnding;
  for I := 0 to Rows.Count - 1 do begin
    Result := Result + '    <contact>' + LineEnding;
    for J := 0 to Rows[I].Cells.Count - 1 do
      if Rows[I].Cells[J] <> '' then
        Result := Result + '      <property name="' + Columns[J] + '">' + ReplaceXmlEntities(Rows[I].Cells[J]) + '</property>' + LineEnding;
    Result := Result + '    </contact>' + LineEnding;
  end;
  Result := Result + '  </contacts>' + LineEnding +
    '</vcard>';
end;

function TTable.GetOutputHtml: string;
var
  I: Integer;
  J: Integer;
begin
  Result := '<html>' + LineEnding +
    '  <head>' + LineEnding +
    '    <title>' + Title + '</title>' + LineEnding +
    '  </head>' + LineEnding +
    '  <body>' + LineEnding +
    '    <table border="1">' + LineEnding;
  // Show header
  Result := Result + '      <tr>' + LineEnding;
  for J := 0 to Columns.Count - 1 do
    Result := Result + '        <th>' + StringReplace(Columns[J], LineEnding, '<br/>', [rfReplaceAll]) + '</th>' + LineEnding;
  Result := Result + '      </tr>' + LineEnding;

  // Show data rows
  for I := 0 to Rows.Count - 1 do begin
    Result := Result + '      <tr>' + LineEnding;
    for J := 0 to Rows[I].Cells.Count - 1 do
      Result := Result + '        <td>' + StringReplace(Rows[I].Cells[J], LineEnding, '<br/>', [rfReplaceAll]) + '</td>' + LineEnding;
    Result := Result + '      </tr>' + LineEnding;
  end;
  Result := Result + '    </table>' + LineEnding +
    '  </body>' + LineEnding +
    '</html>';
end;

function TTable.GetOutputMediaWiki: string;
var
  I: Integer;
  J: Integer;
begin
  Result := '{| class="wikitable sortable"' + LineEnding;

  // Show header
  for J := 0 to Columns.Count - 1 do begin
    if J = 0 then Result := Result + '! ' + Columns[J]
      else Result := Result + ' !! ' + Columns[J];
  end;
  Result := Result + LineEnding + '|-' + LineEnding;

  // Show data rows
  for I := 0 to Rows.Count - 1 do begin
    for J := 0 to Rows[I].Cells.Count - 1 do begin
      if J = 0 then Result := Result + '| ' + Rows[I].Cells[J]
        else Result := Result + ' || ' + Rows[I].Cells[J];
    end;
    if I < Rows.Count - 1 then
      Result := Result + LineEnding + '|-' + LineEnding;
  end;
  Result := Result + LineEnding + '|}';
end;

procedure TTable.GetOutputListView(ListView: TListView);
var
  I: Integer;
  J: Integer;
  ListColumn: TListColumn;
  ListItem: TListItem;
  NewWidth: Integer;
const
  MinWidth = 120;
begin
  ListView.Columns.BeginUpdate;
  try
    ListView.Columns.Clear;
    for I := 0 to Columns.Count - 1 do begin
      ListColumn := ListView.Columns.Add;
      ListColumn.Caption := Columns[I];
      NewWidth := ListView.Width div Columns.Count;
      if NewWidth < MinWidth then NewWidth := MinWidth;
      ListColumn.Width := NewWidth;
    end;
  finally
    ListView.Columns.EndUpdate;
  end;

  ListView.Items.BeginUpdate;
  try
    ListView.Items.Clear;
    for I := 0 to Rows.Count - 1 do begin
      ListItem := ListView.Items.Add;
      for J := 0 to Rows[I].Cells.Count - 1 do begin
        if J = 0 then ListItem.Caption := Rows[I].Cells[J]
          else ListItem.SubItems.Add(Rows[I].Cells[J]);
      end;
    end;
  finally
    ListView.Items.EndUpdate;
  end;
end;

function TTable.GetOutput(OutputFormat: TTableFormat): string;
begin
  case OutputFormat of
    tfExcel: Result := GetOutputTabs;
    tfPlain: Result := GetOutputPlain;
    tfCsv: Result := GetOutputCsv;
    tfHtml: Result := GetOutputHtml;
    tfMediaWiki: Result := GetOutputMediaWiki;
    tfXml: Result := GetOutputXml;
    else Result := '';
  end;
end;

constructor TTable.Create;
begin
  Columns := TStringList.Create;
  Rows := TObjectList<TRow>.Create;
end;

destructor TTable.Destroy;
begin
  FreeAndNil(Rows);
  FreeAndNil(Columns);
  inherited;
end;

procedure TRow.AddCell(Text: string);
begin
  Cells.Add(Text);
end;

constructor TRow.Create;
begin
  Cells := TStringList.Create;
end;

destructor TRow.Destroy;
begin
  FreeAndNil(Cells);
  inherited;
end;

end.

